@@ -2,8 +2,6 @@
 // for details. All rights reserved. Use of this source code is governed by a
 // BSD-style license that can be found in the LICENSE file.
 
-// @dart = 2.6
-
 /// This library defines runtime operations on objects used by the code
 /// generator.
 part of dart._runtime;
@@ -19,9 +17,9 @@ class InvocationImpl extends Invocation {
   final bool isSetter;
   final String failureMessage;
 
-  InvocationImpl(memberName, List<Object> positionalArguments,
+  InvocationImpl(memberName, List<Object?> positionalArguments,
       {namedArguments,
-      List typeArguments,
+      List typeArguments = const [],
       this.isMethod = false,
       this.isGetter = false,
       this.isSetter = false,
@@ -30,9 +28,7 @@ class InvocationImpl extends Invocation {
             isSetter ? _setterSymbol(memberName) : _dartSymbol(memberName),
         positionalArguments = List.unmodifiable(positionalArguments),
         namedArguments = _namedArgsToSymbols(namedArguments),
-        typeArguments = typeArguments == null
-            ? const []
-            : List.unmodifiable(typeArguments.map(wrapType));
+        typeArguments = List.unmodifiable(typeArguments.map(wrapType));
 
   static Map<Symbol, dynamic> _namedArgsToSymbols(namedArgs) {
     if (namedArgs == null) return const {};
@@ -86,22 +82,26 @@ bindCall(obj, name) {
 ///
 /// We need to apply the type arguments both to the function, as well as its
 /// associated function type.
-gbind(f, @rest List typeArgs) {
+gbind(f, @rest List<Object> typeArgs) {
   GenericFunctionType type = JS('!', '#[#]', f, _runtimeType);
   type.checkBounds(typeArgs);
-  // Create a JS wrapper function that will also pass the type arguments, and
-  // tag it with the instantiated function type.
+  // Create a JS wrapper function that will also pass the type arguments.
   var result =
       JS('', '(...args) => #.apply(null, #.concat(args))', f, typeArgs);
+  // Tag the wrapper with the original function to be used for equality checks.
+  JS('', '#["_originalFn"] = #', result, f);
+  JS('', '#["_typeArgs"] = #', result, constList(typeArgs, Object));
+
+  // Tag the wrapper with the instantiated function type.
   return fn(result, type.instantiate(typeArgs));
 }
 
-dloadRepl(obj, field) => dload(obj, replNameLookup(obj, field), false);
+dloadRepl(obj, field) => dload(obj, replNameLookup(obj, field));
 
 // Warning: dload, dput, and dsend assume they are never called on methods
 // implemented by the Object base class as those methods can always be
 // statically resolved.
-dload(obj, field, [@undefined mirrors]) {
+dload(obj, field) {
   if (JS('!', 'typeof # == "function" && # == "call"', obj, field)) {
     return obj;
   }
@@ -115,45 +115,29 @@ dload(obj, field, [@undefined mirrors]) {
     if (hasMethod(type, f)) return bind(obj, f, null);
 
     // Always allow for JS interop objects.
-    if (!JS<bool>('!', '#', mirrors) && isJsInterop(obj)) {
-      return JS('', '#[#]', obj, f);
-    }
+    if (isJsInterop(obj)) return JS('', '#[#]', obj, f);
   }
   return noSuchMethod(obj, InvocationImpl(field, JS('', '[]'), isGetter: true));
 }
 
-// Version of dload that matches legacy mirrors behavior for JS types.
-dloadMirror(obj, field) => dload(obj, field, true);
-
 _stripGenericArguments(type) {
   var genericClass = getGenericClass(type);
   if (genericClass != null) return JS('', '#()', genericClass);
   return type;
 }
 
-// Version of dput that matches legacy Dart 1 type check rules and mirrors
-// behavior for JS types.
-// TODO(jacobr): remove the type checking rules workaround when mirrors based
-// PageLoader code can generate the correct reified generic types.
-dputMirror(obj, field, value) => dput(obj, field, value, true);
+dputRepl(obj, field, value) => dput(obj, replNameLookup(obj, field), value);
 
-dputRepl(obj, field, value) =>
-    dput(obj, replNameLookup(obj, field), value, false);
-
-dput(obj, field, value, [@undefined mirrors]) {
+dput(obj, field, value) {
   var f = _canonicalMember(obj, field);
   trackCall(obj);
   if (f != null) {
     var setterType = getSetterType(getType(obj), f);
     if (setterType != null) {
-      if (JS('!', '#', mirrors))
-        setterType = _stripGenericArguments(setterType);
-      return JS('', '#[#] = #._check(#)', obj, f, setterType, value);
+      return JS('', '#[#] = #.as(#)', obj, f, setterType, value);
     }
     // Always allow for JS interop objects.
-    if (!JS<bool>('!', '#', mirrors) && isJsInterop(obj)) {
-      return JS('', '#[#] = #', obj, f, value);
-    }
+    if (isJsInterop(obj)) return JS('', '#[#] = #', obj, f, value);
   }
   noSuchMethod(
       obj, InvocationImpl(field, JS('', '[#]', value), isSetter: true));
@@ -164,7 +148,7 @@ dput(obj, field, value, [@undefined mirrors]) {
 /// [actuals] and [namedActuals].
 ///
 /// Returns `null` if all checks pass.
-String _argumentErrors(FunctionType type, List actuals, namedActuals) {
+String? _argumentErrors(FunctionType type, List actuals, namedActuals) {
   // Check for too few required arguments.
   int actualsCount = JS('!', '#.length', actuals);
   var required = type.args;
@@ -183,26 +167,46 @@ String _argumentErrors(FunctionType type, List actuals, namedActuals) {
   }
 
   // Check if we have invalid named arguments.
-  Iterable names;
+  Iterable? names;
   var named = type.named;
+  var requiredNamed = type.requiredNamed;
   if (namedActuals != null) {
     names = getOwnPropertyNames(namedActuals);
     for (var name in names) {
-      if (!JS('!', '#.hasOwnProperty(#)', named, name)) {
+      if (!JS<bool>('!', '(#.hasOwnProperty(#) || #.hasOwnProperty(#))', named,
+          name, requiredNamed, name)) {
         return "Dynamic call with unexpected named argument '$name'.";
       }
     }
   }
+  // Verify that all required named parameters are provided an argument.
+  Iterable requiredNames = getOwnPropertyNames(requiredNamed);
+  if (requiredNames.isNotEmpty) {
+    var missingRequired = namedActuals == null
+        ? requiredNames
+        : requiredNames.where((name) =>
+            !JS<bool>('!', '#.hasOwnProperty(#)', namedActuals, name));
+    if (missingRequired.isNotEmpty) {
+      var error = "Dynamic call with missing required named arguments: "
+          "${missingRequired.join(', ')}.";
+      if (!compileTimeFlag('soundNullSafety')) {
+        _nullWarn(error);
+      } else {
+        return error;
+      }
+    }
+  }
   // Now that we know the signature matches, we can perform type checks.
   for (var i = 0; i < requiredCount; ++i) {
-    JS('', '#[#]._check(#[#])', required, i, actuals, i);
+    JS('', '#[#].as(#[#])', required, i, actuals, i);
   }
   for (var i = 0; i < extras; ++i) {
-    JS('', '#[#]._check(#[#])', optionals, i, actuals, i + requiredCount);
+    JS('', '#[#].as(#[#])', optionals, i, actuals, i + requiredCount);
   }
   if (names != null) {
     for (var name in names) {
-      JS('', '#[#]._check(#[#])', named, name, namedActuals, name);
+      JS('', '(#[#] || #[#]).as(#[#])', named, name, requiredNamed, name,
+          namedActuals, name);
     }
   }
   return null;
@@ -250,6 +254,12 @@ Symbol _setterSymbol(name) {
           _toDisplayName(name));
 }
 
+/// Checks for a valid function, receiver and arguments before calling [f].
+///
+/// If passed, [args] and [typeArgs] must be native JavaScript arrays.
+///
+/// NOTE: The contents of [args] may be modified before calling [f]. If it
+/// originated outside of the SDK it must be copied first.
 _checkAndCall(f, ftype, obj, typeArgs, args, named, displayName) =>
     JS('', '''(() => {
   $trackCall($obj);
@@ -260,7 +270,9 @@ _checkAndCall(f, ftype, obj, typeArgs, args, named, displayName) =>
     return $noSuchMethod(originalTarget, new $InvocationImpl.new(
         $displayName, $args, {
           namedArguments: $named,
-          typeArguments: $typeArgs,
+          // Repeated the default value here to avoid passing null from JS to a
+          // non-nullable argument.
+          typeArguments: $typeArgs || [],
           isMethod: true,
           failureMessage: errorMessage
         }));
@@ -298,12 +310,8 @@ _checkAndCall(f, ftype, obj, typeArgs, args, named, displayName) =>
     return $f.apply($obj, $args);
   }
 
-  // TODO(vsm): Remove when we no longer need mirrors metadata.
-  // An array is used to encode annotations attached to the type.
-  if ($ftype instanceof Array) $ftype = $ftype[0];
-
   // Apply type arguments
-  if ($ftype instanceof $GenericFunctionType) {
+  if (${_jsInstanceOf(ftype, GenericFunctionType)}) {
     let formalCount = $ftype.formalCount;
 
     if ($typeArgs == null) {
@@ -331,8 +339,8 @@ _checkAndCall(f, ftype, obj, typeArgs, args, named, displayName) =>
 dcall(f, args, [@undefined named]) => _checkAndCall(
     f, null, JS('', 'void 0'), null, args, named, JS('', 'f.name'));
 
-dgcall(f, typeArgs, args, [@undefined named]) =>
-    _checkAndCall(f, null, JS('', 'void 0'), typeArgs, args, named, 'call');
+dgcall(f, typeArgs, args, [@undefined named]) => _checkAndCall(f, null,
+    JS('', 'void 0'), typeArgs, args, named, JS('', "f.name || 'call'"));
 
 /// Helper for REPL dynamic invocation variants that make a best effort to
 /// enable accessing private members across library boundaries.
@@ -407,52 +415,80 @@ dindex(obj, index) => callMethod(obj, '_get', null, [index], null, '[]');
 dsetindex(obj, index, value) =>
     callMethod(obj, '_set', null, [index, value], null, '[]=');
 
+/// General implementation of the Dart `is` operator.
+///
+/// Some basic cases are handled directly by the `.is` methods that are attached
+/// directly on types, but any query that requires checking subtyping relations
+/// is handled here.
 @notNull
 @JSExportName('is')
 bool instanceOf(obj, type) {
   if (obj == null) {
-    return identical(type, unwrapType(Null)) || _isTop(type);
+    return _equalType(type, Null) ||
+        _isTop(type) ||
+        _jsInstanceOf(type, NullableType);
   }
   return isSubtypeOf(getReifiedType(obj), type);
 }
 
+/// General implementation of the Dart `as` operator.
+///
+/// Some basic cases are handled directly by the `.as` methods that are attached
+/// directly on types, but any query that requires checking subtyping relations
+/// is handled here.
 @JSExportName('as')
-cast(obj, type, @notNull bool isImplicit) {
-  if (obj == null) return obj;
-  var actual = getReifiedType(obj);
-  if (isSubtypeOf(actual, type)) {
+cast(obj, type) {
+  // We hoist the common case where null is checked against another type here
+  // for better performance.
+  if (obj == null && !compileTimeFlag('soundNullSafety')) {
+    // Check the null comparison cache to avoid emitting repeated warnings.
+    _nullWarnOnType(type);
     return obj;
+  } else {
+    var actual = getReifiedType(obj);
+    if (isSubtypeOf(actual, type)) return obj;
   }
-  return castError(obj, type, isImplicit);
+
+  return castError(obj, type);
 }
 
-bool test(bool obj) {
-  if (obj == null) _throwBooleanConversionError();
+bool test(bool? obj) {
+  if (obj == null) throw BooleanConversionAssertionError();
   return obj;
 }
 
 bool dtest(obj) {
-  if (obj is! bool) booleanConversionFailed(obj);
+  // Only throw an AssertionError in weak mode for compatibility. Strong mode
+  // should throw a TypeError.
+  if (obj is! bool)
+    booleanConversionFailed(
+        compileTimeFlag('soundNullSafety') ? obj : test(obj));
   return obj;
 }
 
-void _throwBooleanConversionError() => throw BooleanConversionAssertionError();
-
-void booleanConversionFailed(obj) {
-  var actual = typeName(getReifiedType(test(obj)));
+Never booleanConversionFailed(obj) {
+  var actual = typeName(getReifiedType(obj));
   throw TypeErrorImpl("type '$actual' is not a 'bool' in boolean expression");
 }
 
 asInt(obj) {
-  if (obj == null) return null;
-
+  // Note: null (and undefined) will fail this test.
   if (JS('!', 'Math.floor(#) != #', obj, obj)) {
-    castError(obj, JS('', '#', int), false);
+    if (obj == null && !compileTimeFlag('soundNullSafety')) {
+      _nullWarnOnType(JS('', '#', int));
+      return null;
+    } else {
+      castError(obj, JS('', '#', int));
+    }
   }
   return obj;
 }
 
-/// Checks that `x` is not null or undefined.
+asNullableInt(obj) => obj == null ? null : asInt(obj);
+
+/// Checks for null or undefined and returns [x].
+///
+/// Throws [NoSuchMethodError] when it is null or undefined.
 //
 // TODO(jmesserly): inline this, either by generating it as a function into
 // the module, or via some other pattern such as:
@@ -465,18 +501,46 @@ _notNull(x) {
   return x;
 }
 
+/// Checks for null or undefined and returns [x].
+///
+/// Throws a [TypeError] when [x] is null or undefined (under sound null safety
+/// mode) or emits a runtime warning (otherwise).
+///
+/// This is only used by the compiler when casting from nullable to non-nullable
+/// variants of the same type.
+nullCast(x, type) {
+  if (x == null) {
+    if (!compileTimeFlag('soundNullSafety')) {
+      _nullWarnOnType(type);
+    } else {
+      castError(x, type);
+    }
+  }
+  return x;
+}
+
+/// Checks for null or undefined and returns [x].
+///
+/// Throws a [TypeError] when [x] is null or undefined.
+///
+/// This is only used by the compiler for the runtime null check operator `!`.
+nullCheck(x) {
+  if (x == null) throw TypeErrorImpl("Unexpected null value.");
+  return x;
+}
+
 /// The global constant map table.
-final constantMaps = JS('', 'new Map()');
+final constantMaps = JS<Object>('!', 'new Map()');
 
 // TODO(leafp): This table gets quite large in apps.
 // Keeping the paths is probably expensive.  It would probably
 // be more space efficient to just use a direct hash table with
 // an appropriately defined structural equality function.
-Object _lookupNonTerminal(Object map, Object key) {
+Object _lookupNonTerminal(Object map, Object? key) {
   var result = JS('', '#.get(#)', map, key);
   if (result != null) return result;
   JS('', '#.set(#, # = new Map())', map, key, result);
-  return result;
+  return result!;
 }
 
 Map<K, V> constMap<K, V>(JSArray elements) {
@@ -486,14 +550,14 @@ Map<K, V> constMap<K, V>(JSArray elements) {
     map = _lookupNonTerminal(map, JS('', '#[#]', elements, i));
   }
   map = _lookupNonTerminal(map, K);
-  var result = JS('', '#.get(#)', map, V);
+  Map<K, V>? result = JS('', '#.get(#)', map, V);
   if (result != null) return result;
   result = ImmutableMap<K, V>.from(elements);
   JS('', '#.set(#, #)', map, V, result);
   return result;
 }
 
-final constantSets = JS('', 'new Map()');
+final constantSets = JS<Object>('!', 'new Map()');
 var _immutableSetConstructor;
 
 // We cannot invoke private class constructors directly in Dart.
@@ -509,21 +573,13 @@ Set<E> constSet<E>(JSArray<E> elements) {
   for (var i = 0; i < count; i++) {
     map = _lookupNonTerminal(map, JS('', '#[#]', elements, i));
   }
-  var result = JS('', '#.get(#)', map, E);
+  Set<E>? result = JS('', '#.get(#)', map, E);
   if (result != null) return result;
   result = _createImmutableSet<E>(elements);
   JS('', '#.set(#, #)', map, E, result);
   return result;
 }
 
-bool dassert(value) {
-  if (JS('!', '# != null && #[#] instanceof #', value, value, _runtimeType,
-      AbstractFunctionType)) {
-    value = dcall(value, []);
-  }
-  return dtest(value);
-}
-
 final _value = JS('', 'Symbol("_value")');
 
 ///
@@ -560,7 +616,7 @@ multiKeyPutIfAbsent(map, keys, valueFn) => JS('', '''(() => {
 /// and value of the field.  The final map is
 /// indexed by runtime type, and contains the canonical
 /// version of the object.
-final constants = JS('', 'new Map()');
+final constants = JS('!', 'new Map()');
 
 ///
 /// Canonicalize a constant object.
@@ -619,7 +675,7 @@ constList(elements, elementType) => JS('', '''(() => {
   let value = map.get($elementType);
   if (value) return value;
 
-  ${getGenericClass(JSArray)}($elementType).unmodifiable($elements);
+  ${getGenericClassStatic<JSArray>()}($elementType).unmodifiable($elements);
   map.set($elementType, elements);
   return elements;
 })()''');
@@ -661,7 +717,11 @@ String _toString(obj) {
 String str(obj) {
   if (obj == null) return "null";
   if (obj is String) return obj;
-  return _notNull(JS('!', '#[#]()', obj, extensionSymbol('toString')));
+  final result = JS('', '#[#]()', obj, extensionSymbol('toString'));
+  // TODO(40614): Declare `result` as String once non-nullability is sound.
+  if (result is String) return result;
+  // Since Dart 2.0, `null` is the only other option.
+  throw ArgumentError.value(obj, 'object', "toString method returned 'null'");
 }
 
 // TODO(jmesserly): is the argument type verified statically?
@@ -679,7 +739,7 @@ runtimeType(obj) {
   return obj == null ? Null : JS('', '#[dartx.runtimeType]', obj);
 }
 
-final identityHashCode_ = JS('', 'Symbol("_identityHashCode")');
+final identityHashCode_ = JS<Object>('!', 'Symbol("_identityHashCode")');
 
 /// Adapts a Dart `get iterator` into a JS `[Symbol.iterator]`.
 // TODO(jmesserly): instead of an adaptor, we could compile Dart iterators
@@ -713,15 +773,174 @@ _canonicalMember(obj, name) {
   return name;
 }
 
+/// A map from libraries to a set of import prefixes that have been loaded.
+///
+/// Used to validate deferred library conventions.
+final deferredImports = JS<Object>('!', 'new Map()');
+
 /// Emulates the implicit "loadLibrary" function provided by a deferred library.
 ///
-/// Libraries are not actually deferred in DDC, so this just returns a future
-/// that completes immediately.
-Future loadLibrary() => Future.value();
+/// Libraries are not actually deferred in DDC, so this just records the import
+/// for runtime validation, then returns a future that completes immediately.
+Future loadLibrary(
+    @notNull String enclosingLibrary, @notNull String importPrefix) {
+  var result = JS('', '#.get(#)', deferredImports, enclosingLibrary);
+  if (JS<bool>('', '# === void 0', result)) {
+    JS('', '#.set(#, # = new Set())', deferredImports, enclosingLibrary,
+        result);
+  }
+  JS('', '#.add(#)', result, importPrefix);
+  return Future.value();
+}
 
-/// Defines lazy statics.
-void defineLazy(to, from) {
-  for (var name in getOwnNamesAndSymbols(from)) {
-    defineLazyField(to, name, getOwnPropertyDescriptor(from, name));
+void checkDeferredIsLoaded(
+    @notNull String enclosingLibrary, @notNull String importPrefix) {
+  var loaded = JS('', '#.get(#)', deferredImports, enclosingLibrary);
+  if (JS<bool>('', '# === void 0', loaded) ||
+      JS<bool>('', '!#.has(#)', loaded, importPrefix)) {
+    throwDeferredIsLoadedError(enclosingLibrary, importPrefix);
   }
 }
+
+/// Defines lazy statics.
+///
+/// TODO: Remove useOldSemantics when non-null-safe late static field behavior is
+/// deprecated.
+void defineLazy(to, from, bool useOldSemantics) {
+  for (var name in getOwnNamesAndSymbols(from)) {
+    if (useOldSemantics) {
+      defineLazyFieldOld(to, name, getOwnPropertyDescriptor(from, name));
+    } else {
+      defineLazyField(to, name, getOwnPropertyDescriptor(from, name));
+    }
+  }
+}
+
+/// Defines a lazy static field.
+/// After initial get or set, it will replace itself with a value property.
+// TODO(jmesserly): reusing descriptor objects has been shown to improve
+// performance in other projects (e.g. webcomponents.js ShadowDOM polyfill).
+defineLazyField(to, name, desc) => JS('', '''(() => {
+  const initializer = $desc.get;
+  const final = $desc.set == null;
+  // Tracks if the initializer has been called.
+  let initialized = false;
+  let init = initializer;
+  let value = null;
+  // Tracks if these local variables have been saved so they can be restored
+  // after a hot restart.
+  let savedLocals = false;
+  $desc.get = function() {
+    if (init == null) return value;
+    if (final && initialized) $throwLateInitializationError($name);
+    if (!savedLocals) {
+      // Record the field on first execution so we can reset it later if
+      // needed (hot restart).
+      $resetFields.push(() => {
+        init = initializer;
+        value = null;
+        savedLocals = false;
+        initialized = false;
+      });
+      savedLocals = true;
+    }
+    // Must set before calling init in case it is recursive.
+    initialized = true;
+    try {
+      value = init();
+    } catch (e) {
+      // Reset to false so the initializer can be executed again if the
+      // exception was caught.
+      initialized = false;
+      throw e;
+    }
+    init = null;
+    return value;
+  };
+  $desc.configurable = true;
+  if ($desc.set != null) {
+    $desc.set = function(x) {
+      if (!savedLocals) {
+        $resetFields.push(() => {
+          init = initializer;
+          value = null;
+          savedLocals = false;
+          initialized = false;
+        });
+        savedLocals = true;
+      }
+      init = null;
+      value = x;
+    };
+  }
+  return ${defineProperty(to, name, desc)};
+})()''');
+
+/// Defines a lazy static field with pre-null-safety semantics.
+defineLazyFieldOld(to, name, desc) => JS('', '''(() => {
+  const initializer = $desc.get;
+  let init = initializer;
+  let value = null;
+  // Tracks if these local variables have been saved so they can be restored
+  // after a hot restart.
+  let savedLocals = false;
+  $desc.get = function() {
+    if (init == null) return value;
+    let f = init;
+    init = $throwCyclicInitializationError;
+    if (f === init) f($name); // throw cycle error
+
+    // On the first (non-cyclic) execution, record the field so we can reset it
+    // later if needed (hot restart).
+    if (!savedLocals) {
+      $resetFields.push(() => {
+        init = initializer;
+        value = null;
+        savedLocals = false;
+      });
+      savedLocals = true;
+    }
+
+    // Try to evaluate the field, using try+catch to ensure we implement the
+    // correct Dart error semantics.
+    try {
+      value = f();
+      init = null;
+      return value;
+    } catch (e) {
+      init = null;
+      value = null;
+      throw e;
+    }
+  };
+  $desc.configurable = true;
+  if ($desc.set != null) {
+    $desc.set = function(x) {
+      if (!savedLocals) {
+        $resetFields.push(() => {
+          init = initializer;
+          value = null;
+          savedLocals = false;
+        });
+        savedLocals = true;
+      }
+      init = null;
+      value = x;
+    };
+  }
+  return ${defineProperty(to, name, desc)};
+})()''');
+
+checkNativeNonNull(dynamic variable) {
+  if (_nativeNonNullAsserts && variable == null) {
+    // TODO(srujzs): Add link/patch for instructions to disable in internal
+    // build systems.
+    throw TypeErrorImpl('''
+      Unexpected null value encountered in Dart web platform libraries.
+      This may be a bug in the Dart SDK APIs. If you would like to report a bug
+      or disable this error, you can use the following instructions:
+      https://github.com/dart-lang/sdk/tree/master/sdk/lib/html/doc/NATIVE_NULL_ASSERTIONS.md
+    ''');
+  }
+  return variable;
+}
